# Written by Bram Cohen
# modified for multitracker operation by John Hoffman
# see LICENSE.txt for license information
# Updated and modified for ABC_OKC : Old King Cole

from BitTornado.zurllib import urlopen
from urllib import quote
from BitTornado.utility import exceptionArgsToString, sysencoding, getABCUtility
from urlparse import urlparse
from socket import error as socketerror
from BitTornado.bencode import bdecode
from threading import Thread, Lock
from StringIO import StringIO
from traceback import print_exc
from random import shuffle
from hashlib import sha1
from time import clock, time
try:
    from os import getpid
except ImportError:
    def getpid():
        return 1
    

mapbase64 = '0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz.-'
keys = {}
basekeydata = str(getpid()) + repr(time()) + 'tracker'
ints = (long, int)

def add_key(tracker):
    key = ''
    for i in sha1(basekeydata + tracker).digest()[-6:]:
        key += mapbase64[ord(i) & 0x3F]
    keys[tracker] = key

def get_key(tracker):
    try:
        return "&key=" + keys[tracker]
    except:
        add_key(tracker)
        return "&key=" + keys[tracker]

def check_peers(message, ipv6enabled):
    if type(message) is not dict:
        raise ValueError
    if message.has_key('failure reason'):
        if type(message['failure reason']) is not str:
            raise ValueError
        return
    peers = message.get('peers')
    if type(peers) is list:
        for p in peers:
            if type(p) is not dict:
                raise ValueError
            if type(p.get('ip')) is not str:
                raise ValueError
            port = p.get('port')
            if type(port) not in ints:
                raise ValueError
            if p.has_key('peer id'):
                id = p['peer id']
                if type(id) is not str:
                    raise ValueError
    elif type(peers) is not str or len(peers) % 6 != 0:
        raise ValueError
    if ipv6enabled:
        peers6 = message.get('peers6')
        if peers6 is not None and (type(peers6) is not str or len(peers6) % 18 != 0):
            raise ValueError
    interval = message.get('interval', 1)
    if type(interval) not in ints or interval <= 0:
        raise ValueError
    minint = message.get('min interval', 1)
    if type(minint) not in ints or minint <= 0:
        raise ValueError
    if type(message.get('tracker id', '')) is not str:
        raise ValueError
    npeers = message.get('num peers', 0)
    if type(npeers) not in ints or npeers < 0:
        raise ValueError
    dpeers = message.get('done peers', 0)
    if type(dpeers) not in ints or dpeers < 0:
        raise ValueError
    last = message.get('last', 0)
    if type(last) not in ints or last < 0:
        raise ValueError


class Rerequester:
    def __init__(self, port, myid, infohash, trackerlist, config,
                 sched, errorfunc, excfunc, connect,
                 howmany, amount_left, up, down, upratefunc, downratefunc,
                 doneflag, unpauseflag, seededfunc = None,
                 special = None, trackerokfunc = None, rejectexceptions = []):
        self.sched = sched
        self.errorfunc = errorfunc
        self.excfunc = excfunc
        self.connect = connect
        self.howmany = howmany
        self.amount_left = amount_left
        self.up = up
        self.down = down
        self.upratefunc = upratefunc
        self.downratefunc = downratefunc
        self.doneflag = doneflag
        self.unpauseflag = unpauseflag
        self.seededfunc = seededfunc

        self.ip = config.get('ip')
        self.multihomeipv4 = quote(config.get('multihome_ipv4'), '')
        self.multihomeipv6 = quote(config.get('multihome_ipv6'), '')
        self.minpeers = config['min_peers']
        self.maxpeers = config['max_initiate']
        self.interval = config['rerequest_interval']
        self.timeout = config['http_timeout']
        self.ipv6enabled = config['ipv6_enabled']

        self.trackerlist = []
        for tier in trackerlist:
            if len(tier) > 1:
                shuffle(tier)
            self.trackerlist += [tier]
        self.defaulttracker = config['default_tracker']
        if not self.defaulttracker:
            self.defaulttracker = None

        self.lastsuccessful = ''
        self.rejectedmessage = getABCUtility().lang.get('rejectedtracker') + ' - '

        self.url = ('info_hash=%s&peer_id=%s' % (quote(infohash, ''), quote(myid, '')))
        if not config.get('crypto_allowed'):
            self.url += "&port="
        else:
            self.url += "&supportcrypto=1"
            if not config.get('crypto_only'):
                    self.url += "&port="
            else:
                self.url += "&requirecrypto=1"            
                if not config.get('crypto_stealth'):
                    self.url += "&port="
                else:
                    self.url += "&port=0&cryptoport="
        self.url += str(port)

        seed_id = config.get('dedicated_seed_id')
        if seed_id:
            self.url += '&seed_id='+quote(seed_id, '')
        if self.seededfunc:
            self.url += '&check_seeded=1'

        self.last = None
        self.trackerid = None
        self.announce_interval = 1800
        self.last_failed = True
        self.tryingtoconnect = True
        self.errorcodes = {}
        self.rejectexceptions = rejectexceptions
        # True if one rejected message from trackers contains one string from the list rejectexceptions
        # In this case the torrent is not considered as definitely rejected and will not be stopped
        # if rejection immediately occurs after starting
        self.rejectexception = False
        self.lock = SuccessLock()
        self.special = special
        self.stopped = True
        self.last_announce = 0
        self.first_reqcycle = True
        self.trackerokfunc = trackerokfunc
        # To store the last connection time in a retry cycle for each tracker
        self.trackercnx = {}
        self.lastannouncedtracker = None

    def start(self, event = 0):
        self.stopped = False
        for tl in self.trackerlist:
            for t in tl:
                self.trackercnx[t] = 0
        if self.special is not None:
            self.trackercnx[self.special] = 0
        self.d(event)

    def d(self, event = 3):
        if self.stopped:
            return
        if not self.unpauseflag.isSet():
            # One min interval cycle has elapsed and the torrent is paused
            self.last_failed = True
            self.stopped = True
            return
        if self.special:
            lastcnx = self.trackercnx[self.special]
        elif self.trackercnx:
            lastcnx = max(self.trackercnx.values())
        else:
            lastcnx = 0
        if self.tryingtoconnect:
            extrainterval = 60 - clock() + lastcnx
        else:
            self.first_reqcycle = False
            extrainterval = self.interval - clock() + lastcnx
        if extrainterval > 0:
            self.sched(self.d, extrainterval)
            return
        if not self.trackerlist and self.special is None:
            self.tryingtoconnect = False
            self._d()
        elif self.tryingtoconnect:
            self.announce(event, self._d)    
        elif self.howmany() < self.minpeers or clock() - self.last_announce > self.announce_interval:
            self.tryingtoconnect = True
            self.announce(event, self._d)
        else:
            self._d()

    def _d(self):
        if self.tryingtoconnect:
            # Retry every 60 seconds if never succeeded
            self.sched(self.d, 60)
        else:
            self.first_reqcycle = False
            self.sched(self.d, self.interval)

    def hit(self, event = 3):
        return
        #if not self.unpauseflag.isSet() \
        #   and self.howmany() < self.minpeers:
        #    self.announce(event)

    def announce(self, event = 3, callback = lambda: None, tracker = None, specialurl = None):
        if tracker == 'ext' and specialurl is not None:
            self.special = specialurl
            self.trackercnx[self.special] = 0
        elif tracker == 'int' and self.special is not None:
            specialintrackerlist = False
            for tl in self.trackerlist:
                if self.special in tl:
                    specialintrackerlist = True
                    break
            if not specialintrackerlist and self.trackercnx.has_key(self.special):
                del self.trackercnx[self.special]
            self.special = None
            if not self.trackerlist:
                return
        if not self.unpauseflag.isSet():
            # Don't announce if torrent is paused (but set internal or external tracker parameters)
            return
        s = ('%s&uploaded=%s&downloaded=%s&left=%s' %
            (self.url, str(self.up()), str(self.down()), 
            str(self.amount_left())))
        if self.last is not None:
            s += '&last=' + quote(str(self.last), '')
        if self.trackerid is not None:
            s += '&trackerid=' + quote(str(self.trackerid), '')
        if self.howmany() >= self.maxpeers:
            s += '&numwant=0'
        else:
            s += '&no_peer_id=1&compact=1'
        if event != 3:
            s += '&event=' + ['started', 'completed', 'stopped'][event]
        if event == 2:
            self.stopped = True
        if self.ipv6enabled:
            if self.multihomeipv4:
                s += '&ipv4=' + self.multihomeipv4
            if self.multihomeipv6:
                s += '&ipv6=' + self.multihomeipv6
        self.rerequest(s, callback)

    def rerequest(self, s, callback):
        if not self.lock.isfinished():  # still waiting for prior cycle to complete??
            self.sched(self.rerequest, 5, [s, callback])         # retry in 5 seconds
            return
        self.lock.reset()
        thread = Thread(target = self._rerequest, args = [s, callback, self.special])
        thread.daemon = False
        thread.start()

    def _rerequest(self, s, callback, special):
        try:
            if self.ip:
                ip = getABCUtility().getHostByName(self.ip)
                if ip is None:
                    self.errorcodes['troublecode'] = getABCUtility().lang.get('cantresolve') + self.ip
                    self.sched(self._fail, 0, [callback, special])
                else:
                    s += '&ip=' + ip
            self.errorcodes = {}
            self.rejectexception = False
            now = clock()
            if special is None:
                if self.defaulttracker is not None and self.rerequest_single(self.defaulttracker, s, callback):
                    if not self.last_failed:
                        for tlr in xrange(len(self.trackerlist)):
                            for tr in xrange(len(self.trackerlist[tlr])):
                                if self.trackerlist[tlr][tr] == self.defaulttracker and tr != 0:
                                    del self.trackerlist[tlr][tr]
                                    self.trackerlist[tlr].insert(0, self.defaulttracker)
                    self.defaulttracker = None
                    return
                for tlr in xrange(len(self.trackerlist)):
                    for tr in xrange(len(self.trackerlist[tlr])):
                        tracker = self.trackerlist[tlr][tr]
                        if tracker != self.defaulttracker and self.rerequest_single(tracker, s, callback):
                            if not self.last_failed and tr != 0:
                                del self.trackerlist[tlr][tr]
                                self.trackerlist[tlr].insert(0, tracker)
                            self.defaulttracker = None
                            return
                self.defaulttracker = None
            elif self.rerequest_single(special, s, callback, external = True):
                return
            # No success from any tracker
            self.sched(self._fail, 0, [callback, special])
        except:
            self.exception(callback)

    def _fail(self, callback, special):
        if self.unpauseflag.isSet():
            for f in ['rejected', 'bad_data', 'troublecode']:
                if self.errorcodes.has_key(f):
                    r = self.errorcodes[f]
                    break
            else:
                # All trackers have already been contacted less than self.interval ago and have answered with
                # an error ; so none has been rerequested -> silently fail
                r = ''
            if r:
                self.errorfunc(r, self.first_reqcycle and f == 'rejected' and len(self.errorcodes) == 1 and not self.rejectexception, special is not None)

        self.last_failed = True
        self.lock.give_up()
        self.sched(callback)

    def rerequest_single(self, t, s, callback, external = False):
        l = self.lock.set()
        thread = Thread(target = self._rerequest_single, args = [t, s + get_key(t), l, callback])
        thread.daemon = False
        thread.start()
        self.lock.wait()
        if self.lock.success:
            self.lastsuccessful = t
            self.last_failed = False
            self.trackerokfunc(t, external)
            return True
        if not self.last_failed and self.lastsuccessful == t:
            # If the last tracker hit was successful, and you've just tried the tracker
            # you'd contacted before, don't go any further, just fail silently.
            self.last_failed = True
            self.sched(callback)
            self.lock.give_up()
            return True
        return False

    def _rerequest_single(self, t, s, l, callback):
        try:
            closer = [None]
            def timedout(self = self, l = l, closer = closer):
                if self.lock.trip(l):
                    self.errorcodes['troublecode'] = getABCUtility().lang.get('cantreachtracker') + ' - ' + getABCUtility().lang.get('timeout')
                    self.trackercnx[t] = 0
                    self.lock.unwait(l)
                try:
                    closer[0]()
                except:
                    pass

            self.sched(timedout, self.timeout)

            err = None
            if urlparse(t)[4] != '':
                infohashprefix = '&'
            else:
                infohashprefix = '?'
            self.last_announce = clock()
            try:
                h = urlopen(t + infohashprefix + s)
                if not self.stopped:
                    closer[0] = h.close
                    data = h.read()
            except (IOError, socketerror), e:
                err = getABCUtility().lang.get('cantreachtracker') + ' - ' + exceptionArgsToString(e)
            except:
                err = getABCUtility().lang.get('cantreachtracker')
            try:
                h.close()
            except:
                pass

            if self.stopped:
                if self.lock.trip(l):
                    self.lock.unwait(l)
                return

            if err:        
                if self.lock.trip(l):
                    self.errorcodes['troublecode'] = err
                    self.trackercnx[t] = 0
                    self.lock.unwait(l)
                return

            if data == '':
                if self.lock.trip(l):
                    self.errorcodes['troublecode'] = getABCUtility().lang.get('nodatatracker')
                    self.trackercnx[t] = clock()
                    self.lock.unwait(l)
                return

            try:
                r = bdecode(data, sloppy = 1)
                check_peers(r, self.ipv6enabled)
            except ValueError, e:
                if self.lock.trip(l):
                    self.errorcodes['bad_data'] = getABCUtility().lang.get('baddatatracker') + ' - ' + exceptionArgsToString(e)
                    self.trackercnx[t] = clock()
                    self.lock.unwait(l)
                return

            if r.has_key('failure reason'):
                if self.lock.trip(l):
                    failurereason = r['failure reason']
                    self.errorcodes['rejected'] = self.rejectedmessage + failurereason
                    if not self.rejectexception:
                        for s in self.rejectexceptions:
                            if failurereason.find(s) >= 0:
                                self.rejectexception = True
                                break
                    self.trackercnx[t] = clock()
                    self.lock.unwait(l)
                return

            if self.lock.trip(l, True):     # success!
                self.tryingtoconnect = False
                self.lock.unwait(l)
            else:
                callback = lambda: None     # attempt timed out, don't do a callback

            # even if the attempt timed out, go ahead and process data
            self.sched(self.postrequest, 0, [r, callback, t])
        except:
            self.exception(callback)

    def postrequest(self, r, callback, t):
        if r.has_key('warning message'):
            self.errorfunc(getABCUtility().lang.get('warningtracker') + ' - ' + r['warning message'])
        self.announce_interval = r.get('interval', self.announce_interval)
        self.interval = r.get('min interval', self.interval)
        self.trackerid = r.get('tracker id', self.trackerid)
        self.last = r.get('last')
        p = r['peers']
        peers = []
        if type(p) is str:
            lenpeers = len(p) / 6
        else:
            lenpeers = len(p)
        cflags = r.get('crypto_flags')
        if type(cflags) is not str or len(cflags) != lenpeers:
            cflags = None
        if cflags is None:
            cflags = lenpeers * [None]
        else:
            cflags = [ord(x) for x in cflags]
        if type(p) is str:
            for x in xrange(0, len(p), 6):
                ip = '.'.join([str(ord(i)) for i in p[x:x + 4]])
                port = (ord(p[x + 4]) << 8) + ord(p[x + 5])
                peers.append(((ip, port), 0, cflags[x / 6], None))
        else:
            for i in xrange(lenpeers):
                x = p[i]
                peers.append(((x['ip'].strip(), x['port']),
                             x.get('peer id', 0), cflags[i], None))
        if self.ipv6enabled:
            p6 = r.get('peers6')
            if p6 is not None:
                for x in xrange(0, len(p6), 18):
                    hexpeerinfo = hexlify(p6[x:x + 16])
                    ip = ':'.join([hexpeerinfo[i:i + 4] for i in xrange(0, 32, 4)])
                    port = (ord(p6[x + 16]) << 8) + ord(p6[x + 17])
                    peers.append(((ip, port), 0, None, None))
        ps = len(peers) + self.howmany()
        if ps < self.maxpeers:
            if self.doneflag.isSet():
                if r.get('num peers', 1000) - r.get('done peers', 0) > ps * 1.2:
                    self.last = None
            else:
                if r.get('num peers', 1000) > ps * 1.2:
                    self.last = None
        if self.unpauseflag.isSet():
            if self.seededfunc and r.get('seeded'):
                self.seededfunc()
            elif peers:
                shuffle(peers)
                self.connect(peers)
        callback()

    def exception(self, callback):
        try:
            data = StringIO()
            print_exc(file = data)
            def r(s = data.getvalue().decode(sysencoding), callback = callback):
                if self.excfunc:
                    self.excfunc(s)
                else:
                    print s
                callback()
            self.sched(r)
        except:
            pass


class SuccessLock:
    def __init__(self):
        self.lock = Lock()
        self.pause = Lock()
        self.code = 0L
        self.success = False
        self.finished = True

    def reset(self):
        self.success = False
        self.finished = False

    def set(self):
        self.lock.acquire()
        if not self.pause.locked():
            self.pause.acquire()
        self.first = True
        self.code += 1L
        self.lock.release()
        return self.code

    def trip(self, code, s = False):
        self.lock.acquire()
        try:
            if code == self.code and not self.finished:
                r = self.first
                self.first = False
                if s:
                    self.finished = True
                    self.success = True
                return r
        finally:
            self.lock.release()

    def give_up(self):
        self.lock.acquire()
        self.success = False
        self.finished = True
        self.lock.release()

    def wait(self):
        self.pause.acquire()

    def unwait(self, code):
        if code == self.code and self.pause.locked():
            self.pause.release()

    def isfinished(self):
        self.lock.acquire()
        x = self.finished
        self.lock.release()
        return x    
